def parse_script_args(args):
	"""
	Separate the command line arguments into arguments for pip
	and arguments to Python.

	>>> parse_script_args(['foo', '--', 'bar'])
	(['foo'], ['bar'])

	>>> parse_script_args(['foo', 'bar'])
	(['foo', 'bar'], [])
	"""
	try:
		pivot = args.index('--')
	except ValueError:
		pivot = len(args)
	return args[:pivot], args[pivot+1:]



import os
import subprocess
import sys
import signal
import glob
import itertools


def _read_pth_file(filename):
	root = os.path.dirname(filename)
	return (
		os.path.join(root, path.rstrip())
		for path in open(filename)
		if path.strip()
		and not path.startswith('#')
		and not path.startswith('import ')
	)


def _read_pth_files(target):
	"""
	As .pth files aren't honored except in site dirs,
	read the paths indicated by them.
	"""
	pth_files = glob.glob(os.path.join(target, '*.pth'))
	return itertools.chain.from_iterable(map(_read_pth_file, pth_files))


def _build_env(target):
	"""
	Prepend target and .pth references in target to PYTHONPATH
	"""
	env = dict(os.environ)
	suffix = env.get('PYTHONPATH')
	prefix = target,
	items = itertools.chain(
		prefix,
		_read_pth_files(target),
		(suffix,) if suffix else (),
	)
	joined = os.pathsep.join(items)
	env['PYTHONPATH'] = joined
	return env


def with_path(target, params):
	"""
	Launch Python with target on the path and params
	"""
	def null_handler(signum, frame):
		pass

	signal.signal(signal.SIGINT, null_handler)
	cmd = [sys.executable] + params
	subprocess.Popen(cmd, env=_build_env(target)).wait()


def with_path_overlay(target, params):
	"""
	Overlay Python with target on the path and params
	"""
	cmd = [sys.executable] + params
	os.execve(sys.executable, cmd, _build_env(target))



import pkg_resources

from rwt import deps

def test_entry_points():
	"""
	Ensure entry points are visible after making packages visible
	"""
	with deps.on_sys_path('jaraco.mongodb'):
		eps = pkg_resources.iter_entry_points('pytest11')
		assert list(eps), "Entry points not found"



import sys

from . import deps
from . import commands
from . import launch
from . import scripts


if __name__ == '__main__':
	pip_args, params = commands.parse_script_args(sys.argv[1:])
	pip_args.extend(scripts.DepsReader.search(params))
	with deps.load(*pip_args) as home:
		launch.with_path(home, params)



"""
This script demonstrates how RWT facilitates the
simple execution of complex tasks with their
dependencies.

Run this example with ``python -m rwt -- $script``.

It creates a MongoDB instance, and then runs some
assertions against it.

MongoDB must be installed to a typical location or
available on PATH.

As it uses jaraco.mongodb, set MONGODB_HOME to
specify the MongoDB version to use for the ephemeral
instance.

Running this script with RWT leaves no trace of its
execution, other than adding packages to the pip
cache (if available).
"""

__requires__ = [
	'pytest',
	'jaraco.mongodb>=3.10',
]

if __name__ == '__main__':
	# invoke pytest on this script
	import pytest
	import sys
	sys.exit(pytest.main(sys.argv))

import random
import itertools

import pytest
from jaraco.mongodb.testing import assert_covered


@pytest.fixture
def docs_in_db(mongodb_instance):
	"""
	Install 100 records with random numbers
	"""
	conn = mongodb_instance.get_connection()
	coll = conn.test_db.test_coll
	coll.drop()
	coll.create_index('number')
	n_records = 100
	for n in itertools.islice(itertools.count(), n_records):
		doc = dict(
			number=random.randint(0, 2**32-1),
			value='some value',
		)
		conn.test_db.test_coll.insert(doc)
	return coll


def test_records(docs_in_db, mongodb_instance):
	"Test 100 records are present and query is covered"
	# load all the numbers to ensure the index is in RAM
	_hint = 'number_1'
	_filter = {'number': {'$gt': 0}}
	_projection = {'number': True, '_id': False}
	cur = docs_in_db.find(filter=_filter, projection=_projection).hint(_hint)
	assert_covered(cur)

	# consume the cursor for good measure
	docs = list(cur)
	assert len(docs) == 100



from __future__ import unicode_literals

import textwrap
import sys
import subprocess

from rwt import scripts


def test_pkg_imported(tmpdir):
	"""
	Create a script that loads cython and ensure it runs.
	"""
	body = textwrap.dedent("""
		import cython
		print("Successfully imported cython")
		""").lstrip()
	script_file = tmpdir / 'script'
	script_file.write_text(body, 'utf-8')
	pip_args = ['cython']
	cmd = [sys.executable, '-m', 'rwt'] + pip_args + ['--', str(script_file)]

	out = subprocess.check_output(cmd, universal_newlines=True)
	assert 'Successfully imported cython' in out


class TestDepsReader:
	def test_reads_files_with_attribute_assignment(self):
		script = textwrap.dedent('''
			__requires__=['foo']
			x.a = 'bar'
			''')
		assert scripts.DepsReader(script).read() == ['foo']

	def test_reads_files_with_multiple_assignment(self):
		script = textwrap.dedent('''
			__requires__=['foo']
			x, a = [a, x]
			''')
		assert scripts.DepsReader(script).read() == ['foo']



from __future__ import print_function

import sys
import contextlib
import subprocess
import tempfile
import shutil


@contextlib.contextmanager
def _update_working_set():
	"""
	Update the master working_set to include these new packages.

	TODO: would be better to use an officially-supported API,
	but no suitable API is apparent.
	"""
	try:
		pkg_resources = sys.modules['pkg_resources']
		if not hasattr(pkg_resources, '_initialize_master_working_set'):
			exec(_init_ws_patch, vars(pkg_resources))
		pkg_resources._initialize_master_working_set()
	except KeyError:
		# it's unnecessary to re-initialize when it hasn't
		# yet been initialized.
		pass
	yield


@contextlib.contextmanager
def load(*args):
	target = tempfile.mkdtemp(prefix='rwt-')
	cmdline = subprocess.list2cmdline(args)
	print("Loading requirements using", cmdline)
	cmd = (
		sys.executable,
		'-m', 'pip',
			'install',
			'-q',
			'-t', target,
	) + args
	subprocess.check_call(cmd)
	try:
		yield target
	finally:
		shutil.rmtree(target)


@contextlib.contextmanager
def on_sys_path(*args):
	"""
	Install dependencies via args to pip and ensure they have precedence
	on sys.path.
	"""
	with load(*args) as target:
		sys.path.insert(0, target)
		try:
			with _update_working_set():
				yield target
		finally:
			sys.path.remove(target)


# from setuptools 19.6.2
_init_ws_patch = '''
def _initialize_master_working_set():
    """
    Prepare the master working set and make the ``require()``
    API available.

    This function has explicit effects on the global state
    of pkg_resources. It is intended to be invoked once at
    the initialization of this module.

    Invocation by other packages is unsupported and done
    at their own risk.
    """
    working_set = WorkingSet._build_master()
    _declare_state('object', working_set=working_set)

    require = working_set.require
    iter_entry_points = working_set.iter_entry_points
    add_activation_listener = working_set.subscribe
    run_script = working_set.run_script
    # backward compatibility
    run_main = run_script
    # Activate all distributions already on sys.path, and ensure that
    # all distributions added to the working set in the future (e.g. by
    # calling ``require()``) will get activated as well.
    add_activation_listener(lambda dist: dist.activate())
    working_set.entries=[]
    # match order
    list(map(working_set.add_entry, sys.path))
    globals().update(locals())
'''



import os
import sys
import ast
import tokenize
import itertools


if sys.version_info < (3,):
	filter = itertools.ifilter


class DepsReader:
	"""
	Given a Python script, read the dependencies from the
	indicated variable (default __requires__). Does not
	execute the script, so expects the var_name to be
	assigned a static list of strings.
	"""
	def __init__(self, script):
		self.script = script

	@classmethod
	def load(cls, script_path):
		with open(script_path) as stream:
			return cls(stream.read())

	@classmethod
	def try_read(cls, script_path):
		"""
		Attempt to load the dependencies from the script,
		but return an empty list if unsuccessful.
		"""
		try:
			reader = cls.load(script_path)
			return reader.read()
		except Exception:
			return []

	@classmethod
	def search(cls, params):
		"""
		Given a (possibly-empty) series of parameters to a
		Python interpreter, return any dependencies discovered
		in a script indicated in the parameters. Only honor the
		first file found.
		"""
		files = filter(os.path.isfile, params)
		return cls.try_read(next(files, None))

	def read(self, var_name='__requires__'):
		"""
		>>> DepsReader("__requires__=['foo']").read()
		['foo']
		"""
		mod = ast.parse(self.script)
		node, = (
			node
			for node in mod.body
			if isinstance(node, ast.Assign)
			and len(node.targets) == 1
			and isinstance(node.targets[0], ast.Name)
			and node.targets[0].id == var_name
		)
		return ast.literal_eval(node.value)


def run(cmdline):
	"""
	Execute the script as if it had been invoked naturally.
	"""
	namespace = dict()
	filename = cmdline[0]
	namespace['__file__'] = filename
	namespace['__name__'] = '__main__'
	sys.argv[:] = cmdline

	open_ = getattr(tokenize, 'open', open)
	script = open_(filename).read()
	norm_script = script.replace('\\r\\n', '\\n')
	code = compile(norm_script, filename, 'exec')
	exec(code, namespace)
